@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997, 1998, 1999 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -38,7 +38,7 @@
 
 #include "kdc_locl.h"
 
-RCSID("$Id: kerberos5.c,v 1.78 1997/11/12 01:27:17 joda Exp $");
+RCSID("$Id: kerberos5.c,v 1.103 1999/07/03 02:26:52 assar Exp $");
 
 #define MAX_TIME ((time_t)((1U << 31) - 1))
 
@@ -76,6 +76,40 @@ find_padata(KDC_REQ *req, int *start, int type)
     return NULL;
 }
 
+#if 0
+
+static krb5_error_code
+find_keys(hdb_entry *client, 
+	  hdb_entry *server, 
+	  Key **ckey, 
+	  krb5_enctype *cetype,
+	  Key **skey,
+	  krb5_enctype *setype,
+	  unsigned *etypes, 
+	  unsigned num_etypes)
+{
+    int i;
+    krb5_error_code ret;
+    for(i = 0; i < num_etypes; i++) {
+	if(client){
+	    ret = hdb_enctype2key(context, client, etypes[i], ckey);
+	    if(ret)
+		continue;
+	}
+	if(server){
+	    ret = hdb_enctype2key(context, server, etypes[i], skey);
+	    if(ret)
+		continue;
+	}
+	if(etype)
+	    *cetype = *setype = etypes[i];
+	return 0;
+    }
+    return KRB5KDC_ERR_ETYPE_NOSUPP;
+}
+
+#else
+
 static krb5_error_code
 find_etype(hdb_entry *princ, unsigned *etypes, unsigned len, 
 	   Key **key, int *index)
@@ -83,17 +117,21 @@ find_etype(hdb_entry *princ, unsigned *etypes, unsigned len,
     int i;
     krb5_error_code ret = -1;
     for(i = 0; i < len ; i++)
-	if((ret = hdb_etype2key(context, princ, etypes[i], key)) == 0)
+	if((ret = hdb_enctype2key(context, princ, etypes[i], key)) == 0)
 	    break;
     if(index) *index = i;
     return ret;
 }
 
 static krb5_error_code
-find_keys(hdb_entry *client, hdb_entry *server, 
-	  Key **ckey, krb5_enctype *cetype,
-	  Key **skey, krb5_enctype *setype, krb5_keytype *sess_ktype,
-	  unsigned *etypes, unsigned num_etypes)
+find_keys(hdb_entry *client,
+	  hdb_entry *server, 
+	  Key **ckey,
+	  krb5_enctype *cetype,
+	  Key **skey,
+	  krb5_enctype *setype, 
+	  unsigned *etypes,
+	  unsigned num_etypes)
 {
     int i;
     krb5_error_code ret;
@@ -108,43 +146,29 @@ find_keys(hdb_entry *client, hdb_entry *server,
     }
 
     if(server){
-	/* find sesion key type */
+	/* find server key */
 	ret = find_etype(server, etypes, num_etypes, skey, NULL);
 	if(ret){
 	    kdc_log(0, "Server has no support for etypes");
 	    return KRB5KDC_ERR_ETYPE_NOSUPP;
 	}
-	*sess_ktype = (*skey)->key.keytype;
-    }
-    if(server){
-	/* find server key */
-	*skey = NULL;
-#define is_better(x, y) ((x) > (y))
-	for(i = 0; i < server->keys.len; i++){
-	    if(*skey == NULL || is_better(server->keys.val[i].key.keytype, 
-					  (*skey)->key.keytype))
-		*skey = &server->keys.val[i];
-	}
-	if(*skey == NULL){
-	    kdc_log(0, "No key found for server");
-	    return KRB5KDC_ERR_NULL_KEY;
-	}
-	ret = krb5_keytype_to_etype(context, (*skey)->key.keytype, setype);
-	if(ret)
-	    return ret;
+	*setype = (*skey)->key.keytype;
     }
     return 0;
 }
+#endif
 
 static krb5_error_code
 encode_reply(KDC_REP *rep, EncTicketPart *et, EncKDCRepPart *ek, 
-	     krb5_enctype setype, int skvno, EncryptionKey *skey,
-	     krb5_enctype cetype, int ckvno, EncryptionKey *ckey,
+	     krb5_enctype etype, 
+	     int skvno, EncryptionKey *skey,
+	     int ckvno, EncryptionKey *ckey,
 	     krb5_data *reply)
 {
     unsigned char buf[8192]; /* XXX The data could be indefinite */
     size_t len;
     krb5_error_code ret;
+    krb5_crypto crypto;
 
     ret = encode_EncTicketPart(buf + sizeof(buf) - 1, sizeof(buf), et, &len);
     if(ret) {
@@ -153,15 +177,20 @@ encode_reply(KDC_REP *rep, EncTicketPart *et, EncKDCRepPart *ek,
 	return ret;
     }
     
+
+    krb5_crypto_init(context, skey, etype, &crypto);
+
     krb5_encrypt_EncryptedData(context, 
+			       crypto,
+			       KRB5_KU_TICKET,
 			       buf + sizeof(buf) - len,
 			       len,
-			       setype,
 			       skvno,
-			       skey,
 			       &rep->ticket.enc_part);
+
+    krb5_crypto_destroy(context, crypto);
     
-    if(rep->msg_type == krb_as_rep)
+    if(rep->msg_type == krb_as_rep && !encode_as_rep_as_tgs_rep)
 	ret = encode_EncASRepPart(buf + sizeof(buf) - 1, sizeof(buf), 
 				  ek, &len);
     else
@@ -172,18 +201,27 @@ encode_reply(KDC_REP *rep, EncTicketPart *et, EncKDCRepPart *ek,
 		krb5_get_err_text(context, ret));
 	return ret;
     }
-    krb5_encrypt_EncryptedData(context,
-			       buf + sizeof(buf) - len,
-			       len,
-			       cetype,
-			       ckvno,
-			       ckey,
-			       &rep->enc_part);
-	
-    if(rep->msg_type == krb_as_rep)
+    krb5_crypto_init(context, ckey, 0, &crypto);
+    if(rep->msg_type == krb_as_rep) {
+	krb5_encrypt_EncryptedData(context,
+				   crypto,
+				   KRB5_KU_AS_REP_ENC_PART,
+				   buf + sizeof(buf) - len,
+				   len,
+				   ckvno,
+				   &rep->enc_part);
 	ret = encode_AS_REP(buf + sizeof(buf) - 1, sizeof(buf), rep, &len);
-    else
+    } else {
+	krb5_encrypt_EncryptedData(context,
+				   crypto,
+				   KRB5_KU_TGS_REP_ENC_PART_SESSION,
+				   buf + sizeof(buf) - len,
+				   len,
+				   ckvno,
+				   &rep->enc_part);
 	ret = encode_TGS_REP(buf + sizeof(buf) - 1, sizeof(buf), rep, &len);
+    }
+    krb5_crypto_destroy(context, crypto);
     if(ret) {
 	kdc_log(0, "Failed to encode KDC-REP: %s", 
 		krb5_get_err_text(context, ret));
@@ -193,23 +231,198 @@ encode_reply(KDC_REP *rep, EncTicketPart *et, EncKDCRepPart *ek,
     return 0;
 }
 
+static int
+realloc_method_data(METHOD_DATA *md)
+{
+    PA_DATA *pa;
+    pa = realloc(md->val, (md->len + 1) * sizeof(*md->val));
+    if(pa == NULL)
+	return ENOMEM;
+    md->val = pa;
+    md->len++;
+    return 0;
+}
+
+static krb5_error_code
+get_pa_etype_info(METHOD_DATA *md, hdb_entry *client)
+{
+    krb5_error_code ret = 0;
+    int i;
+    ETYPE_INFO pa;
+    unsigned char *buf;
+    size_t len;
+    
+
+    pa.len = client->keys.len;
+    pa.val = malloc(pa.len * sizeof(*pa.val));
+    if(pa.val == NULL)
+	return ENOMEM;
+    for(i = 0; i < client->keys.len; i++) {
+	pa.val[i].etype = client->keys.val[i].key.keytype;
+	ALLOC(pa.val[i].salttype);
+	if(client->keys.val[i].salt){
+#if 0
+	    if(client->keys.val[i].salt->type == hdb_pw_salt)
+		*pa.val[i].salttype = 0; /* or 1? or NULL? */
+	    else if(client->keys.val[i].salt->type == hdb_afs3_salt)
+		*pa.val[i].salttype = 2;
+	    else {
+		free_ETYPE_INFO(&pa);
+		kdc_log(0, "unknown salt-type: %d", 
+			client->keys.val[i].salt->type);
+		return KRB5KRB_ERR_GENERIC;
+	    }
+	    /* according to `the specs', we can't send a salt if
+	       we have AFS3 salted key, but that requires that you
+	       *know* what cell you are using (e.g by assuming
+	       that the cell is the same as the realm in lower
+	       case) */
+#else
+	    *pa.val[i].salttype = client->keys.val[i].salt->type;
+#endif
+	    krb5_copy_data(context, &client->keys.val[i].salt->salt,
+			   &pa.val[i].salt);
+	} else {
+#if 0
+	    *pa.val[i].salttype = 1; /* or 0 with salt? */
+#else
+	    *pa.val[i].salttype = pa_pw_salt;
+#endif
+	    pa.val[i].salt = NULL;
+	}
+    }
+    len = length_ETYPE_INFO(&pa);
+    buf = malloc(len);
+    if (buf) {
+	free_ETYPE_INFO(&pa);
+	return ret;
+    }
+    ret = encode_ETYPE_INFO(buf + len - 1, len, &pa, &len);
+    free_ETYPE_INFO(&pa);
+    if(ret) {
+	free(buf);
+	return ret;
+    }
+    ret = realloc_method_data(md);
+    if(ret) {
+	free(buf);
+	return ret;
+    }
+    md->val[md->len - 1].padata_type = pa_etype_info;
+    md->val[md->len - 1].padata_value.length = len;
+    md->val[md->len - 1].padata_value.data = buf;
+    return 0;
+}
+
+static int
+check_flags(hdb_entry *client, const char *client_name,
+	    hdb_entry *server, const char *server_name,
+	    krb5_boolean is_as_req)
+{
+    if(client != NULL) {
+	/* check client */
+	if (client->flags.invalid) {
+	    kdc_log(0, "Client (%s) has invalid bit set", client_name);
+	    return KRB5KDC_ERR_POLICY;
+	}
+	
+	if(!client->flags.client){
+	    kdc_log(0, "Principal may not act as client -- %s", 
+		    client_name);
+	    return KRB5KDC_ERR_POLICY;
+	}
+	
+	if (client->valid_start && *client->valid_start > kdc_time) {
+	    kdc_log(0, "Client not yet valid -- %s", client_name);
+	    return KRB5KDC_ERR_CLIENT_NOTYET;
+	}
+	
+	if (client->valid_end && *client->valid_end < kdc_time) {
+	    kdc_log(0, "Client expired -- %s", client_name);
+	    return KRB5KDC_ERR_NAME_EXP;
+	}
+	
+	if (client->pw_end && *client->pw_end < kdc_time
+	    && !server->flags.change_pw) {
+	    kdc_log(0, "Client's key has expired -- %s", client_name);
+	    return KRB5KDC_ERR_KEY_EXPIRED;
+	}
+    }
+
+    /* check server */
+    
+    if (server != NULL) {
+	if (server->flags.invalid) {
+	    kdc_log(0, "Server has invalid flag set -- %s", server_name);
+	    return KRB5KDC_ERR_POLICY;
+	}
+
+	if(!server->flags.server){
+	    kdc_log(0, "Principal may not act as server -- %s", 
+		    server_name);
+	    return KRB5KDC_ERR_POLICY;
+	}
+
+	if(!is_as_req && server->flags.initial) {
+	    kdc_log(0, "AS-REQ is required for server -- %s", server_name);
+	    return KRB5KDC_ERR_POLICY;
+	}
+
+	if (server->valid_start && *server->valid_start > kdc_time) {
+	    kdc_log(0, "Server not yet valid -- %s", server_name);
+	    return KRB5KDC_ERR_SERVICE_NOTYET;
+	}
+
+	if (server->valid_end && *server->valid_end < kdc_time) {
+	    kdc_log(0, "Server expired -- %s", server_name);
+	    return KRB5KDC_ERR_SERVICE_EXP;
+	}
+
+	if (server->pw_end && *server->pw_end < kdc_time) {
+	    kdc_log(0, "Server's key has expired -- %s", server_name);
+	    return KRB5KDC_ERR_KEY_EXPIRED;
+	}
+    }
+    return 0;
+}
+
+static krb5_boolean
+check_addresses(HostAddresses *addresses, struct sockaddr *from)
+{
+    krb5_error_code ret;
+    krb5_address addr;
+    
+    if(check_ticket_addresses == 0)
+	return TRUE;
+
+    if(addresses == NULL)
+	return allow_null_ticket_addresses;
+    
+    ret = krb5_sockaddr2address (from, &addr);
+    if(ret)
+	return FALSE;
+
+    return krb5_address_search(context, &addr, addresses);
+}
+
 krb5_error_code
 as_rep(KDC_REQ *req, 
        krb5_data *reply,
-       const char *from)
+       const char *from,
+       struct sockaddr *from_addr)
 {
     KDC_REQ_BODY *b = &req->req_body;
     AS_REP rep;
     KDCOptions f = b->kdc_options;
     hdb_entry *client = NULL, *server = NULL;
     krb5_enctype cetype, setype;
-    krb5_keytype sess_ktype;
     EncTicketPart et;
     EncKDCRepPart ek;
     krb5_principal client_princ, server_princ;
     char *client_name, *server_name;
     krb5_error_code ret = 0;
     const char *e_text = NULL;
+    krb5_crypto crypto;
 
     Key *ckey, *skey;
 
@@ -243,31 +456,6 @@ as_rep(KDC_REQ *req,
 	goto out;
     }
 
-    if (client->valid_start && *client->valid_start > kdc_time) {
-	kdc_log(0, "Client not yet valid -- %s", client_name);
-	ret = KRB5KDC_ERR_CLIENT_NOTYET;
-	goto out;
-    }
-
-    if (client->valid_end && *client->valid_end < kdc_time) {
-	kdc_log(0, "Client expired -- %s", client_name);
-	ret = KRB5KDC_ERR_NAME_EXP;
-	goto out;
-    }
-
-    if(!client->flags.client){
-	ret = KRB5KDC_ERR_POLICY;
-	kdc_log(0, "Principal may not act as client -- %s", 
-		client_name);
-	goto out;
-    }
-
-    if (client->flags.invalid) {
-	ret = KRB5KDC_ERR_POLICY;
-	kdc_log(0, "Client (%s) has invalid bit set", client_name);
-	goto out;
-    }
-
     server = db_fetch(server_princ);
 
     if(server == NULL){
@@ -276,43 +464,9 @@ as_rep(KDC_REQ *req,
 	goto out;
     }
 
-    if (server->valid_start && *server->valid_start > kdc_time) {
-	kdc_log(0, "Server not yet valid -- %s", server_name);
-	ret = KRB5KDC_ERR_SERVICE_NOTYET;
+    ret = check_flags(client, client_name, server, server_name, TRUE);
+    if(ret)
 	goto out;
-    }
-
-    if (server->valid_end && *server->valid_end < kdc_time) {
-	kdc_log(0, "Server expired -- %s", server_name);
-	ret = KRB5KDC_ERR_SERVICE_EXP;
-	goto out;
-    }
-
-    if(!server->flags.server){
-	ret = KRB5KDC_ERR_POLICY;
-	kdc_log(0, "Principal (%s) may not act as server -- %s", 
-		server_name, client_name);
-	goto out;
-    }
-
-    if (server->flags.invalid) {
-	ret = KRB5KDC_ERR_POLICY;
-	kdc_log(0, "Server (%s) has invalid bit set", server_name);
-	goto out;
-    }
-
-    if (server->pw_end && *server->pw_end < kdc_time) {
-	ret = KRB5KDC_ERR_KEY_EXPIRED;
-	kdc_log(0, "Servers key has expired", server_name);
-	goto out;
-    }
-
-    if (client->pw_end && *client->pw_end < kdc_time
-	&& !server->flags.change_pw) {
-	ret = KRB5KDC_ERR_KEY_EXPIRED;
-	kdc_log(0, "Client (%s)'s key has expired", client_name);
-	goto out;
-    }
 
     memset(&et, 0, sizeof(et));
     memset(&ek, 0, sizeof(ek));
@@ -343,7 +497,7 @@ as_rep(KDC_REQ *req,
 		goto out;
 	    }
 	    
-	    ret = hdb_etype2key(context, client, enc_data.etype, &pa_key);
+	    ret = hdb_enctype2key(context, client, enc_data.etype, &pa_key);
 	    if(ret){
 		e_text = "No key matches pa-data";
 		ret = KRB5KDC_ERR_PREAUTH_FAILED;
@@ -353,10 +507,13 @@ as_rep(KDC_REQ *req,
 		continue;
 	    }
 	    
+	    krb5_crypto_init(context, &pa_key->key, 0, &crypto);
 	    ret = krb5_decrypt_EncryptedData (context,
+					      crypto,
+					      KRB5_KU_PA_ENC_TIMESTAMP,
 					      &enc_data,
-					      &pa_key->key,
 					      &ts_data);
+	    krb5_crypto_destroy(context, crypto);
 	    free_EncryptedData(&enc_data);
 	    if(ret){
 		e_text = "Failed to decrypt PA-DATA";
@@ -398,28 +555,34 @@ as_rep(KDC_REQ *req,
 	    e_text = NULL;
 	    goto out;
 	}
-    }else if (require_preauth || client->flags.require_preauth) {
-	/* XXX check server->flags.require_preauth? */
+    }else if (require_preauth || client->flags.require_preauth || server->flags.require_preauth) {
 	METHOD_DATA method_data;
-	PA_DATA pa_data;
-	u_char buf[16];
+	PA_DATA *pa;
+	unsigned char *buf;
 	size_t len;
 	krb5_data foo_data;
-	
-    use_pa:
-	method_data.len = 1;
-	method_data.val = &pa_data;
 
-	pa_data.padata_type         = pa_enc_timestamp;
-	pa_data.padata_value.length = 0;
-	pa_data.padata_value.data   = NULL;
+    use_pa: 
+	method_data.len = 0;
+	method_data.val = NULL;
+
+	ret = realloc_method_data(&method_data);
+	pa = &method_data.val[method_data.len-1];
+	pa->padata_type		= pa_enc_timestamp;
+	pa->padata_value.length	= 0;
+	pa->padata_value.data	= NULL;
+
+	ret = get_pa_etype_info(&method_data, client); /* XXX check ret */
 	
-	encode_METHOD_DATA(buf + sizeof(buf) - 1,
-			   sizeof(buf),
+	len = length_METHOD_DATA(&method_data);
+	buf = malloc(len);
+	encode_METHOD_DATA(buf + len - 1,
+			   len,
 			   &method_data,
 			   &len);
+	free_METHOD_DATA(&method_data);
 	foo_data.length = len;
-	foo_data.data   = buf + sizeof(buf) - len;
+	foo_data.data   = buf;
 	
 	ret = KRB5KDC_ERR_PREAUTH_REQUIRED;
 	krb5_mk_error(context,
@@ -430,32 +593,27 @@ as_rep(KDC_REQ *req,
 		      server_princ,
 		      0,
 		      reply);
-	
+	free(buf);
 	kdc_log(0, "No PA-ENC-TIMESTAMP -- %s", client_name);
+	ret = 0;
 	goto out2;
     }
-
-
-
-    ret = find_keys(client, server, &ckey, &cetype, &skey, &setype, 
-		    &sess_ktype, b->etype.val, b->etype.len);
-    if(ret)
+    
+    ret = find_keys(client, server, &ckey, &cetype, &skey, &setype,
+		    b->etype.val, b->etype.len);
+    if(ret) {
+	kdc_log(0, "Server/client has no support for etypes");
 	goto out;
+    }
 	
     {
-	char *cet, *set, *skt;
-	krb5_etype_to_string(context, cetype, &cet);
-	if(cetype != setype)
-	    krb5_etype_to_string(context, setype, &set);
-	krb5_keytype_to_string(context, sess_ktype, &skt);
-	if(cetype != setype)
-	    kdc_log(5, "Using %s/%s/%s", cet, set, skt);
-	else
-	    kdc_log(5, "Using %s/%s", cet, skt);
+	char *cet;
+	char *set;
+	krb5_enctype_to_string(context, cetype, &cet);
+	krb5_enctype_to_string(context, setype, &set);
+	kdc_log(5, "Using %s/%s", cet, set);
 	free(cet);
-	if(cetype != setype)
-	    free(set);
-	free(skt);
+	free(set);
     }
     
 
@@ -505,7 +663,14 @@ as_rep(KDC_REQ *req,
 	goto out;
     }
 
-    krb5_generate_random_keyblock(context, sess_ktype, &et.key);
+    /* check for valid set of addresses */
+    if(!check_addresses(b->addresses, from_addr)) {
+	ret = KRB5KRB_AP_ERR_BADADDR;
+	kdc_log(0, "Bad address list requested -- %s", client_name);
+	goto out;
+    }
+
+    krb5_generate_random_keyblock(context, setype, &et.key);
     copy_PrincipalName(b->cname, &et.cname);
     copy_Realm(&b->realm, &et.crealm);
     
@@ -561,7 +726,7 @@ as_rep(KDC_REQ *req,
 	ALLOC(et.caddr);
 	copy_HostAddresses(b->addresses, et.caddr);
     }
-
+    
     copy_EncryptionKey(&et.key, &ek.key);
 
     /* The MIT ASN.1 library (obviously) doesn't tell lengths encoded
@@ -595,12 +760,12 @@ as_rep(KDC_REQ *req,
     ek.nonce = b->nonce;
     if (client->valid_end || client->pw_end) {
 	ALLOC(ek.key_expiration);
-	if (client->valid_end)
+	if (client->valid_end) {
 	    if (client->pw_end)
 		*ek.key_expiration = min(*client->valid_end, *client->pw_end);
 	    else
 		*ek.key_expiration = *client->valid_end;
-	else
+	} else
 	    *ek.key_expiration = *client->pw_end;
     } else
 	ek.key_expiration = NULL;
@@ -624,13 +789,12 @@ as_rep(KDC_REQ *req,
 
     set_salt_padata (&rep.padata, ckey->salt);
     ret = encode_reply(&rep, &et, &ek, setype, server->kvno, &skey->key,
-		       cetype, client->kvno, &ckey->key, reply);
+		       client->kvno, &ckey->key, reply);
     free_EncTicketPart(&et);
     free_EncKDCRepPart(&ek);
     free_AS_REP(&rep);
 out:
     if(ret){
-	/* XXX should just return protocol errors */
 	krb5_mk_error(context,
 		      ret,
 		      e_text,
@@ -669,7 +833,7 @@ check_tgs_flags(KDC_REQ_BODY *b, EncTicketPart *tgt, EncTicketPart *et)
 	    kdc_log(0, "Bad request to validate ticket");
 	    return KRB5KDC_ERR_BADOPTION;
 	}
-	if(*tgt->starttime < kdc_time){
+	if(*tgt->starttime > kdc_time){
 	    kdc_log(0, "Early request to validate ticket");
 	    return KRB5KRB_AP_ERR_TKT_NYV;
 	}
@@ -744,6 +908,7 @@ check_tgs_flags(KDC_REQ_BODY *b, EncTicketPart *tgt, EncTicketPart *et)
 	}
 	et->flags.renewable = 1;
 	ALLOC(et->renew_till);
+	fix_time(&b->rtime);
 	*et->renew_till = *b->rtime;
     }
     if(f.renew){
@@ -757,7 +922,7 @@ check_tgs_flags(KDC_REQ_BODY *b, EncTicketPart *tgt, EncTicketPart *et)
 	    old_life -= *tgt->starttime;
 	else
 	    old_life -= tgt->authtime;
-	et->endtime = *et->starttime + old_life;
+	et->endtime = min(*b->till, *et->starttime + old_life);
     }	    
     
     /* checks for excess flags */
@@ -825,6 +990,7 @@ static krb5_error_code
 tgs_make_reply(KDC_REQ_BODY *b, 
 	       EncTicketPart *tgt, 
 	       EncTicketPart *adtkt, 
+	       AuthorizationData *auth_data,
 	       hdb_entry *server, 
 	       hdb_entry *client, 
 	       krb5_principal client_principal, 
@@ -837,20 +1003,33 @@ tgs_make_reply(KDC_REQ_BODY *b,
     EncTicketPart et;
     KDCOptions f = b->kdc_options;
     krb5_error_code ret;
-    krb5_enctype setype;
+    krb5_enctype etype;
     Key *skey;
     EncryptionKey *ekey;
-    krb5_keytype sess_ktype;
     
-    ret = find_keys(NULL, server, NULL, NULL, &skey, &setype, 
-		    &sess_ktype, b->etype.val, b->etype.len);
-    if(ret)
-	return ret;
-    if(adtkt)
+    if(adtkt) {
+	int i;
+	krb5_keytype kt;
 	ekey = &adtkt->key;
-    else
+	for(i = 0; i < b->etype.len; i++){
+	    ret = krb5_enctype_to_keytype(context, b->etype.val[i], &kt);
+	    if(ret)
+		continue;
+	    if(adtkt->key.keytype == kt)
+		break;
+	}
+	if(i == b->etype.len)
+	    return KRB5KDC_ERR_ETYPE_NOSUPP;
+	etype = b->etype.val[i];
+    }else{
+	ret = find_keys(NULL, server, NULL, NULL, &skey, &etype, 
+			b->etype.val, b->etype.len);
+	if(ret) {
+	    kdc_log(0, "Server has no support for etypes");
+	    return ret;
+	}
 	ekey = &skey->key;
-    ret = krb5_keytype_to_etype(context, ekey->keytype, &setype);
+    }
     
     memset(&rep, 0, sizeof(rep));
     memset(&et, 0, sizeof(et));
@@ -938,8 +1117,9 @@ tgs_make_reply(KDC_REQ_BODY *b,
     et.flags.hw_authent = tgt->flags.hw_authent;
 	    
     /* XXX Check enc-authorization-data */
+    et.authorization_data = auth_data;
 
-    krb5_generate_random_keyblock(context, sess_ktype, &et.key);
+    krb5_generate_random_keyblock(context, etype, &et.key);
     et.crealm = tgt->crealm;
     et.cname = tgt->cname;
 	    
@@ -966,8 +1146,8 @@ tgs_make_reply(KDC_REQ_BODY *b,
        CAST session key. Should the DES3 etype be added to the
        etype list, even if we don't want a session key with
        DES3? */
-    ret = encode_reply(&rep, &et, &ek, setype, adtkt ? 0 : server->kvno, ekey,
-		       cetype, 0, &tgt->key, reply);
+    ret = encode_reply(&rep, &et, &ek, etype, adtkt ? 0 : server->kvno, ekey,
+		       0, &tgt->key, reply);
 out:
     free_TGS_REP(&rep);
     free_TransitedEncoding(&et.transited);
@@ -983,12 +1163,14 @@ out:
 
 static krb5_error_code
 tgs_check_authenticator(krb5_auth_context ac,
-			KDC_REQ_BODY *b, krb5_keyblock *key)
+			KDC_REQ_BODY *b, 
+			krb5_keyblock *key)
 {
     krb5_authenticator auth;
     size_t len;
     unsigned char buf[8192];
     krb5_error_code ret;
+    krb5_crypto crypto;
     
     krb5_auth_getauthenticator(context, ac, &auth);
     if(auth->cksum == NULL){
@@ -996,15 +1178,23 @@ tgs_check_authenticator(krb5_auth_context ac,
 	ret = KRB5KRB_AP_ERR_INAPP_CKSUM;
 	goto out;
     }
-    if (!krb5_checksum_is_keyed(auth->cksum->cksumtype)
-	|| !krb5_checksum_is_collision_proof(auth->cksum->cksumtype)) {
+    /*
+     * according to RFC1510 it doesn't need to be keyed,
+     * but according to the latest draft it needs to.
+     */
+    if (
+#if 0
+!krb5_checksum_is_keyed(context, auth->cksum->cksumtype)
+	||
+#endif
+ !krb5_checksum_is_collision_proof(context, auth->cksum->cksumtype)) {
 	kdc_log(0, "Bad checksum type in authenticator: %d", 
 		auth->cksum->cksumtype);
 	ret =  KRB5KRB_AP_ERR_INAPP_CKSUM;
 	goto out;
     }
 		
-    /* XXX */
+    /* XXX should not re-encode this */
     ret = encode_KDC_REQ_BODY(buf + sizeof(buf) - 1, sizeof(buf),
 			      b, &len);
     if(ret){
@@ -1012,9 +1202,14 @@ tgs_check_authenticator(krb5_auth_context ac,
 		krb5_get_err_text(context, ret));
 	goto out;
     }
-    ret = krb5_verify_checksum(context, buf + sizeof(buf) - len, len,
-			       key,
+    krb5_crypto_init(context, key, 0, &crypto);
+    ret = krb5_verify_checksum(context,
+			       crypto,
+			       KRB5_KU_TGS_REQ_AUTH_CKSUM,
+			       buf + sizeof(buf) - len, 
+			       len,
 			       auth->cksum);
+    krb5_crypto_destroy(context, crypto);
     if(ret){
 	kdc_log(0, "Failed to verify checksum: %s", 
 		krb5_get_err_text(context, ret));
@@ -1037,7 +1232,8 @@ is_krbtgt(PrincipalName *p)
 static Realm
 find_rpath(Realm r)
 {
-    const char *new_realm = krb5_config_get_string(context->cf, 
+    const char *new_realm = krb5_config_get_string(context,
+						   NULL,
 						   "libdefaults", 
 						   "capath", 
 						   r, 
@@ -1048,9 +1244,10 @@ find_rpath(Realm r)
 
 static krb5_error_code
 tgs_rep2(KDC_REQ_BODY *b,
-	 PA_DATA *pa_data,
+	 PA_DATA *tgs_req,
 	 krb5_data *reply,
-	 const char *from)
+	 const char *from,
+	 struct sockaddr *from_addr)
 {
     krb5_ap_req ap_req;
     krb5_error_code ret;
@@ -1058,17 +1255,20 @@ tgs_rep2(KDC_REQ_BODY *b,
     krb5_auth_context ac = NULL;
     krb5_ticket *ticket = NULL;
     krb5_flags ap_req_options;
+    krb5_flags verify_ap_req_flags;
     const char *e_text = NULL;
+    krb5_crypto crypto;
 
-    hdb_entry *krbtgt;
+    hdb_entry *krbtgt = NULL;
     EncTicketPart *tgt;
     Key *tkey;
     krb5_enctype cetype;
     krb5_principal cp = NULL;
     krb5_principal sp = NULL;
+    AuthorizationData *auth_data = NULL;
 
     memset(&ap_req, 0, sizeof(ap_req));
-    ret = krb5_decode_ap_req(context, &pa_data->padata_value, &ap_req);
+    ret = krb5_decode_ap_req(context, &tgs_req->padata_value, &ap_req);
     if(ret){
 	kdc_log(0, "Failed to decode AP-REQ: %s", 
 		krb5_get_err_text(context, ret));
@@ -1076,6 +1276,7 @@ tgs_rep2(KDC_REQ_BODY *b,
     }
     
     if(!is_krbtgt(&ap_req.ticket.sname)){
+	/* XXX check for ticket.sname == req.sname */
 	kdc_log(0, "PA-DATA is not a ticket-granting ticket");
 	ret = KRB5KDC_ERR_POLICY; /* ? */
 	goto out2;
@@ -1098,28 +1299,39 @@ tgs_rep2(KDC_REQ_BODY *b,
     
     if(ap_req.ticket.enc_part.kvno && 
        *ap_req.ticket.enc_part.kvno != krbtgt->kvno){
-	kdc_log(0, "Ticket kvno = %d, DB kvno = %d", 
+	char *p;
+
+	krb5_unparse_name (context, princ, &p);
+	kdc_log(0, "Ticket kvno = %d, DB kvno = %d (%s)", 
 		*ap_req.ticket.enc_part.kvno,
-		krbtgt->kvno);
+		krbtgt->kvno,
+		p);
+	free (p);
 	ret = KRB5KRB_AP_ERR_BADKEYVER;
 	goto out2;
     }
 
-    ret = hdb_etype2key(context, krbtgt, ap_req.ticket.enc_part.etype, &tkey);
+    ret = hdb_enctype2key(context, krbtgt, ap_req.ticket.enc_part.etype, &tkey);
     if(ret){
 	char *str;
-	krb5_etype_to_string(context, ap_req.ticket.enc_part.etype, &str);
+	krb5_enctype_to_string(context, ap_req.ticket.enc_part.etype, &str);
 	kdc_log(0, "No server key found for %s", str);
 	free(str);
 	ret = KRB5KRB_AP_ERR_BADKEYVER;
 	goto out2;
     }
     
+    if (b->kdc_options.validate)
+	verify_ap_req_flags = KRB5_VERIFY_AP_REQ_IGNORE_INVALID;
+    else
+	verify_ap_req_flags = 0;
+
     ret = krb5_verify_ap_req(context,
 			     &ac,
 			     &ap_req,
 			     princ,
 			     &tkey->key,
+			     verify_ap_req_flags,
 			     &ap_req_options,
 			     &ticket);
 			     
@@ -1136,6 +1348,54 @@ tgs_rep2(KDC_REQ_BODY *b,
 
     ret = tgs_check_authenticator(ac, b, &tgt->key);
 
+    if (b->enc_authorization_data) {
+	krb5_keyblock *subkey;
+	krb5_data ad;
+	ret = krb5_auth_con_getremotesubkey(context,
+					    ac,
+					    &subkey);
+	if(ret){
+	    kdc_log(0, "Failed to get remote subkey: %s", 
+		    krb5_get_err_text(context, ret));
+	    goto out2;
+	}
+	if(subkey == NULL){
+	    ret = krb5_auth_con_getkey(context, ac, &subkey);
+	    if(ret) {
+		kdc_log(0, "Failed to get session key: %s", 
+			krb5_get_err_text(context, ret));
+		goto out2;
+	    }
+	}
+	if(subkey == NULL){
+	    kdc_log(0, "Failed to get key for enc-authorization-data");
+	    ret = KRB5KRB_AP_ERR_BAD_INTEGRITY; /* ? */
+	    goto out2;
+	}
+	krb5_crypto_init(context, subkey, 0, &crypto);
+	ret = krb5_decrypt_EncryptedData (context,
+					  crypto,
+					  KRB5_KU_TGS_REQ_AUTH_DAT_SUBKEY,
+					  b->enc_authorization_data,
+					  &ad);
+	krb5_crypto_destroy(context, crypto);
+	if(ret){
+	    kdc_log(0, "Failed to decrypt enc-authorization-data");
+	    ret = KRB5KRB_AP_ERR_BAD_INTEGRITY; /* ? */
+	    goto out2;
+	}
+	krb5_free_keyblock(context, subkey);
+	ALLOC(auth_data);
+	ret = decode_AuthorizationData(ad.data, ad.length, auth_data, NULL);
+	if(ret){
+	    free(auth_data);
+	    auth_data = NULL;
+	    kdc_log(0, "Failed to decode authorization data");
+	    ret = KRB5KRB_AP_ERR_BAD_INTEGRITY; /* ? */
+	    goto out2;
+	}
+    }
+
     krb5_auth_con_free(context, ac);
 
     if(ret){
@@ -1180,12 +1440,12 @@ tgs_rep2(KDC_REQ_BODY *b,
 		ret = KRB5KDC_ERR_S_PRINCIPAL_UNKNOWN;
 		goto out;
 	    }
-	    ret = hdb_etype2key(context, uu, t->enc_part.etype, &tkey);
+	    ret = hdb_enctype2key(context, uu, t->enc_part.etype, &tkey);
 	    if(ret){
 		ret = KRB5KDC_ERR_ETYPE_NOSUPP; /* XXX */
 		goto out;
 	    }
-	    ret = krb5_decrypt_ticket(context, t, &tkey->key, &adtkt);
+	    ret = krb5_decrypt_ticket(context, t, &tkey->key, &adtkt, 0);
 
 	    if(ret)
 		goto out;
@@ -1240,6 +1500,9 @@ tgs_rep2(KDC_REQ_BODY *b,
 	}
 #endif
 
+	ret = check_flags(client, cpn, server, spn, FALSE);
+	if(ret)
+	    goto out;
 
 	if((b->kdc_options.validate || b->kdc_options.renew) && 
 	   !krb5_principal_compare(context, 
@@ -1249,10 +1512,24 @@ tgs_rep2(KDC_REQ_BODY *b,
 	    ret = KRB5KDC_ERR_SERVER_NOMATCH;
 	    goto out;
 	}
+
+	/* check for valid set of addresses */
+	if(!check_addresses(tgt->caddr, from_addr)) {
+	    ret = KRB5KRB_AP_ERR_BADADDR;
+	    kdc_log(0, "Request from wrong address");
+	    goto out;
+	}
 	
-	ret = tgs_make_reply(b, tgt, 
+	ret = tgs_make_reply(b, 
+			     tgt, 
 			     b->kdc_options.enc_tkt_in_skey ? &adtkt : NULL, 
-			     server, client, cp, krbtgt, cetype, reply);
+			     auth_data,
+			     server, 
+			     client, 
+			     cp, 
+			     krbtgt, 
+			     cetype, 
+			     reply);
 	
     out:
 	free(spn);
@@ -1285,6 +1562,10 @@ out2:
 	free(ticket);
     }
     free_AP_REQ(&ap_req);
+    if(auth_data){
+	free_AuthorizationData(auth_data);
+	free(auth_data);
+    }
 
     if(krbtgt){
 	hdb_free_entry(context, krbtgt);
@@ -1297,11 +1578,12 @@ out2:
 krb5_error_code
 tgs_rep(KDC_REQ *req, 
 	krb5_data *data,
-	const char *from)
+	const char *from,
+	struct sockaddr *from_addr)
 {
     krb5_error_code ret;
-    int i;
-    PA_DATA *pa_data = NULL;
+    int i = 0;
+    PA_DATA *tgs_req = NULL;
 
     if(req->padata == NULL){
 	ret = KRB5KDC_ERR_PREAUTH_REQUIRED; /* XXX ??? */
@@ -1309,18 +1591,15 @@ tgs_rep(KDC_REQ *req,
 	goto out;
     }
     
-    for(i = 0; i < req->padata->len; i++)
-	if(req->padata->val[i].padata_type == pa_tgs_req){
-	    pa_data = &req->padata->val[i];
-	    break;
-	}
-    if(pa_data == NULL){
+    tgs_req = find_padata(req, &i, pa_tgs_req);
+
+    if(tgs_req == NULL){
 	ret = KRB5KDC_ERR_PADATA_TYPE_NOSUPP;
 	
 	kdc_log(0, "TGS-REQ from %s without PA-TGS-REQ", from);
 	goto out;
     }
-    ret = tgs_rep2(&req->req_body, pa_data, data, from);
+    ret = tgs_rep2(&req->req_body, tgs_req, data, from, from_addr);
 out:
     if(ret && data->data == NULL){
 	krb5_mk_error(context,
